from matplotlib import pyplot as plt
from cartopy import crs as ccrs
import cartopy.feature as cfeature
import numpy as np
from matplotlib import colors as mcolors
from scipy import ndimage as ndi
import xarray as xr
from pyproj import Geod
from shapely.geometry import Polygon
from shapely.ops import orient
from enstools.feature.util.data_utils import pb_str_to_datetime
import matplotlib

binary_grey = mcolors.ListedColormap(['white', 'grey'])

############### PLOTTING FUNCTIONS ###############

def plot_timestep(ds, file_name):
    """
    Plots storms in the dataset ds to the given file name.
    The dataset ds must only contain one timestep.

    Parameters
    ----------
    ds: xr.Dataset
    file_name: str
    """

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(11, 4), subplot_kw=dict(projection=ccrs.PlateCarree()))

    levels = np.linspace(97000, 104000, 14)

    ds.storm_areas.plot.contourf(levels=[0.5, 1000], cmap=binary_grey, add_colorbar=False) # , transform=ccrs.NorthPolarStereo())
    ds.msl.plot.contour(levels=levels, vmin=0, extend='max', cmap='Blues')

    ax.coastlines()
    ax.add_feature(cfeature.BORDERS.with_scale('50m'))

    print("Save to " + file_name)
    plt.savefig(file_name, format='png')

    plt.figure().clear()
    plt.close()
    plt.cla()
    plt.clf()


def plot_track(track, filename, storm_areas):
    """
    Plots the given storm track.

    Parameters
    ----------
    track: pb2.Track
        The track as generated by generate_tracks() and located as one element of set.tracks
    filename: str
        The file name
    storm_areas: xr.DataArray
        The storm areas DataArray to add the areas on top.
    """

    nodes = [edge.parent for edge in track.edges]
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(9, 9), subplot_kw=dict(projection=ccrs.PlateCarree()))
    extent = [-50, 30, 40, 85]

    min_time = pb_str_to_datetime(nodes[0].time).timestamp()
    max_time = pb_str_to_datetime(nodes[-1].time).timestamp()
    cmap = matplotlib.cm.get_cmap('rainbow')
    cmap_v = matplotlib.cm.get_cmap('viridis')

    color_wgts = np.linspace(0.0, 1.0, len(nodes))
    colors = ['red', 'yellow', 'green', 'blue', 'purple']

    lats = []
    lons = []
    press = []
    for node_idx, node in enumerate(nodes):
        obj = node.object
        id = obj.id
        time_d = pb_str_to_datetime(node.time).timestamp()

        # current storm area to 1
        storm_areas_t = storm_areas.sel(time=node.time)
        binary_current_area = xr.where(storm_areas_t == id, 1, 0)

        time_weight = (1.0 -(time_d - min_time) / (max_time - min_time) if max_time > min_time else 1.0)
        binary_current_area.plot.contour(levels=[0.5], colors=[cmap(time_weight)])

        lat = obj.properties.min_pressure_pos.lat
        lon = obj.properties.min_pressure_pos.lon
        pres = obj.properties.min_pressure

        lats.append(lat)
        lons.append(lon)
        press.append(pres)

    colors = [cmap_v((pres - 960.0) / (1000.0 - 960.0)) for pres in press]

    ax.coastlines()
    ax.add_feature(cfeature.BORDERS.with_scale('50m'))
    ax.set_extent(extent, crs=ccrs.PlateCarree())

    plt.scatter(x=lons, y=lats,
                s=40,
                c=colors,
                alpha=1,
                transform=ccrs.PlateCarree())

    figure_name = 'track' + '.png' # .replace(':', '_')
    plt.title(nodes[0].time + " - " + nodes[-1].time)
    print("Plot to " + str(filename))

    plt.savefig(filename, format='png')

    plt.figure().clear()
    plt.close()
    plt.cla()
    plt.clf()


############### IDENTIFICATION FUNCTIONS ###############

def get_min_pressure_hPa(storm_area, msl_data, return_pos=False):
    """
    Get the minimum pressure in hPa of the given storm area.

    Parameters
    ----------
    storm_area: xr.DataArray
        The DataArray of the binary storm field.
    msl_data: xr.DataArray
        The DataArray of mean sea level pressure.
    return_pos: bool
        if True also returns the position of lowest pressure.

    Returns
    -------
    Lowest pressure in hPa, and if return_pos=True also lon and lat of that pressure.
    """

    wh = msl_data.data[storm_area.data]
    min_pressure_pa = np.amin(wh)
    if not return_pos:
        return min_pressure_pa / 100.0 # hPa
    else:
        # position in masked version
        masked_msl_data = msl_data.where(storm_area)
        wh_loc = masked_msl_data.where(masked_msl_data == min_pressure_pa, drop=True).squeeze()
        # lat and lon
        longitude = np.atleast_1d(wh_loc.longitude.data)[0].item()
        latitude = np.atleast_1d(wh_loc.latitude.data)[0].item()

        return min_pressure_pa / 100.0, longitude, latitude

def get_msl_areas_by_hPa_thr(msl_data, hPa):
    """
    Get the areas with lower pressure then given threshold as a binary field.

    Parameters
    ----------
    msl_data: xr.DataArray
        The DataArray of mean sea level pressure.
    hPa: float
        The pressure threshold.

    Returns
    -------
    A binary DataArray based on the given maximum threshold for pressure.
    """
    return dataset < hPa * 100.0

def split_areas(areas):
    """
    Split disjunct areas in a binary DataArray into multiple DataArrays.

    Parameters
    ----------
    areas: xr.DataArray
        The binary DataArray to be split.

    Returns
    -------
    A list of DataArrays with the same shape as the input but each only containing True values of one single area.
    """
    # split dataarray areas into list of cohesive regions (DAs)
    lp_area_lab = xr.zeros_like(areas, dtype=int)

    lp_area_lab.data, num_features = ndi.label(areas)
    a_list = []
    for i in range(1, num_features + 1):
        a_list.append(lp_area_lab == i)

    return a_list


def get_storm_size_km2(storm_da):
    """
    Get the storm size in km².

    Parameters
    ----------
    storm_da: xr.DataArray
        The binary DataArray, which is set to True for each point belonging to the storm.

    Returns
    -------
    The "True" area of the DataArray
    """

    # create contour
    res = storm_da.plot.contour(levels=[0.5, 1000])
    paths = res.collections[0].get_paths()

    if len(paths) > 1:
        print("Multi path?") # TODO
    path = paths[0]

    node_list = path.vertices

    polygon = Polygon(node_list)
    geod = Geod(ellps="WGS84")
    poly_area, poly_perimeter = geod.geometry_area_perimeter(orient(polygon))

    return poly_area / 1000000.0 # in km^2

def has_local_minimum(storm_area, msl_data):
    """
    Check whether the storm area has a local minimum. It has a local minimum,
    iff the minimum itself is not along the data boundary.

    Parameters
    ----------
    storm_area: xr.DataArray
        The binary DataArray of the storm area.
    msl_data: xr.DataArray
        The DataArray of the mean sea level pressure data.

    Returns
    -------
    True if storm has a local minimum in the area.
    """

    all_msl = msl_data.data[storm_area.data]
    min_all_data = np.amin(all_msl)

    inner_area = ndi.binary_erosion(storm_area.data)
    inner_msl = msl_data.data[inner_area]

    if inner_msl.size == 0:
        return False # too small

    min_inner_data = np.amin(inner_msl)
    return min_all_data == min_inner_data # true if min not on border


def get_storm_amplitude_hPa(storm_area, msl_data):
    """
    Get the amplitude of the storm.
    This is a measure between the minimum and average pressure within the storm area.

    Parameters
    ----------
    storm_area: xr.DataArray
        The binary storm area.
    msl_data: xr.DataArray
        The mean sea level pressure data.

    Returns
    -------
    The amplitude as measure mean-min in hPa
    """

    all_msl = msl_data.data[storm_area.data]
    min_p = np.amin(all_msl)
    avg_p = np.mean(all_msl)
    ampl_p = avg_p - min_p

    return ampl_p / 100.0

